with stringpack; use stringpack;
with Ada.Tags; use Ada.Tags;

package body Linda is

    protected body Linda_Env is

        entry Output (T: in Tuple_Type) when true
        is
            I: Tuplespace_Iterator_Type;
            NT, RT: Tuple_Type := T;
            B: Blocker_Access;
            Only_Reading: boolean;
        begin
            -- Zum Vergleich einen formalen Blocker einfuegen
            Insert_First (NT, Blocker_Parameter_Type'(true, null, false));
            -- Solange, bis ein auf Input wartender Task gefunden wurde
            loop
                -- einen wartenden Task finden
                I := Find_Matching (Waiting, NT);
                -- Das Tupel mit dem des wartenden Tasks binden
                RT := Bind (Value (I), NT);
                -- Den wartenden Task aus Warteraum entfernen
                Delete (I);
                Only_Reading := Blocker_Parameter_Type (Value (First (RT))).Only_Reading;
                -- Den Blocker entnehmen
                B := Blocker_Parameter_Type (Value (First (RT))).B;
                Delete_First (RT);
                -- Den wartenden Task erloesen
                B.all.Release (RT);
            exit when not Only_Reading;
            end loop;
        exception
            -- Wenn kein wartender Task gefunden wurde
            when No_Tuple_Matches
            -- dann das Tupel in den Tupelraum eintragen
            =>   Insert_First (TupleSpace, T);
        end Output;

        entry Input (T: in out Tuple_Type) when true
        is
            I: Tuplespace_Iterator_Type;
            B: Blocker_Access;
            NT: Tuple_Type := T;
        begin
            -- Ein passendes Tupel im Tupelraum suchen
            I := Find_Matching (TupleSpace, T);
            -- Die Tupel binden
            T := Bind (Value (I), T);
            -- Das gefundene Tupel aus dem Tupelraum entfernen
            Delete (I);
        exception
            -- wenn kein passendes Tupel im Tupelraum gefunden wurde
            when No_Tuple_Matches
            =>   -- einen Blocker erzeugen um den aufrufenden Task warten zu lassen
                 B := new Blocker_Type;
                 -- Den Blocker in das Tupel eintragen
                 Insert_First (NT, Blocker_Parameter_Type'(false, B, false));
                 -- Das Tupel in den Warteraum stellen
                 Insert_First (Waiting, NT);
                 -- Und den Task endlich warten lassen
                 requeue B.all.Wait;
        end Input;

        entry Read (T: in out Tuple_Type) when true
        is
            I: Tuplespace_Iterator_Type;
            B: Blocker_Access;
            NT: Tuple_Type := T;
        begin
            -- Ein passendes Tupel im Tupelraum suchen
            I := Find_Matching (TupleSpace, T);
            -- Die Tupel binden
            T := Bind (Value (I), T);
        exception
            when No_Tuple_Matches
            =>   -- einen Blocker erzeugen um den aufrufenden Task warten zu lassen
                 B := new Blocker_Type;
                 -- Den Blocker in das Tupel eintragen (Achtung: Only_Reading)
                 Insert_First (NT, Blocker_Parameter_Type'(false, B, true));
                 -- Das Tupel in den Warteraum stellen
                 Insert_First (Waiting, NT);
                 -- Und den Task endlich warten lassen
                 requeue B.all.Wait;
                 print ("Ende Read notuplematch");
        end Read;

    end Linda_Env;

    -- Die Linda-Primitive tauchen sofort in den geschuetzten Bereich ein
    procedure Output (T: Tuple_Type) is
    begin
        Linda_Env.Output (T);
    end Output;

    procedure Input  (T: in out Tuple_Type) is
    begin
        Linda_Env.Input (T);
    end Input;

    procedure Read (T: in out Tuple_Type) is
    begin
        Linda_Env.Read (T);
    end Read;

    -- Zwei Parameter sind gleich, wenn sie gleichen Inhalt haben
    function "=" (Left, Right: Char_Parameter_Type) return boolean is
    begin
        return Left.Char = Right.Char;
    end "=";

    function "=" (Left, Right: Int_Parameter_Type) return boolean is
    begin
        return Left.Int = Right.Int;
    end "=";

    -- Der Blockerparameter wird eh nie verglichen
    function "=" (Left, Right: Blocker_Parameter_Type) return boolean is
    begin
        return false;
    end "=";

end Linda;
