with Tuple, Tuplespace, Blocker;
use Tuple, Tuplespace, Blocker;

package Linda is
    -- Die Lindaprimitive
    procedure Output (T: Tuple_Type);
    procedure Input  (T: in out Tuple_Type);
    procedure Read (T: in out Tuple_Type);

    -- Die Tupelmanipulationstypen und -funktionen
    subtype Parameter_Type is Tuple.Parameter_Type;
    subtype Tuple_Type is Tuple.Tuple_Type;
    subtype Tuplespace_Iterator_Type is Tuplespace.Iterator_Type;
    subtype Iterator_Type is Tuple.Iterator_Type;
    procedure Add (T: Tuple_Type; P: Parameter_Type'Class) renames Insert_Last;
    procedure Empty (T: in out Tuple_Type) renames Tuple.Empty;
    function First (T: Tuple_Type) return Iterator_Type renames Tuple.First;
    function Value (I: Iterator_Type) return Parameter_Type'Class renames Tuple.Value;
    function Succ (I: Iterator_Type) return Iterator_Type renames Tuple.Succ;

    -- Hier werden alle moeglichen Parameter-Typen zur Verfuegung gestellt
    -- Sollte jemand andere benoetigen, bitte selbst erzeugen!

    -- Integer:
    type Int_Parameter_Type
    is   new Parameter_Type
         with record
             Int: integer;
         end record;

    -- Character:
    type Char_Parameter_Type
    is   new Parameter_Type
         with record
             Char: character;
         end record;

private

    -- Zwei Tupelraeume. Einer fuer die ausgegebenen Tupel.
    -- Der andere fuer die wartenden Prozesse.
    -- Vor gleichzeitigem Zugriff geschuetzt.
    protected Linda_Env is
        entry Output (T: in Tuple_Type);
        entry Input  (T: in out Tuple_Type);
        entry Read   (T: in out Tuple_Type);
    private
        Tuplespace, Waiting: Tuplespace_Type;
    end Linda_Env;

    -- Die "="-Funktion muss ueberladen werden
    function "=" (Left, Right: Char_Parameter_Type) return boolean;
    function "=" (Left, Right: Int_Parameter_Type) return boolean;

    -- Ein Taskblocker muss im Waiting-Tupelraum eingetragen werden.
    type Blocker_Access is access Blocker_Type;
    type Blocker_Parameter_Type
    is   new Parameter_Type
         with record
             B: Blocker_Access;
             Only_Reading: boolean := false;
         end record;

    function "=" (Left, Right: Blocker_Parameter_Type) return boolean;

end Linda;
