-- Package Tuple

-- Parameter: ein Parameter besteht aus einem Item (Typ beliebig) und aus
--            einem Schalter, der angibt, ob der Parameter formal oder
--            aktuell ist. Bei formalen Parameter hat Item keinen Inhalt.

-- Parameter-Satz: besteht aus einem Parameter und einem Zeiger auf den
--            naechsten Parameter

-- Tupel:     das Tupel ist eine Liste von Parametern.
--

with Ada.Finalization; use Ada.Finalization;

package Tuple is

    type Parameter_Type              -- Klasse zum Ableiten mit verschie-
    is   abstract tagged record               -- denen Typen;
             Formal: boolean;        -- Enthaelt Schalter formal ja/nein
    end record;
    function "=" (Left, Right: Parameter_Type) return boolean is abstract;

    type Tuple_Type is private;      -- Vordefinition Tupel
    type Iterator_Type is private;   -- Vordefinition Tupelverwalter

    function First (T: Tuple_Type) return Iterator_Type;            -- Erster Param.
    function Succ (I: Iterator_Type) return Iterator_Type;          -- naechster Par.
    function Value (I: Iterator_Type) return Parameter_Type'Class;  -- Wert aktueller
    function  End_Of_Tuple (I: Iterator_Type) return boolean;  -- Ende?
    procedure Insert_First (T: Tuple_Type; P: Parameter_Type'Class);-- vorne dazu
    procedure Delete_First (T: Tuple_Type);                    -- vorne weg
    procedure Insert_Last (T: Tuple_Type; P: Parameter_Type'Class); -- hinten dazu
    procedure Empty (T: in out Tuple_Type);                    -- alle loeschen

    -- diese Funktionen vergleichen 2 Tupel; Bind bindet ausserdem die
    -- aktuellen an die formalen Parameter und liefert ein neues Tupel,
    -- das nur mehr aktuelle Parameter enthaelt, zurueck
    function Match (A, B: Tuple_Type) return boolean;
    function Bind (A, B: Tuple_Type) return Tuple_Type;
    Tuple_Does_Not_Match: exception;

private
                                           -- Pointer auf belieb. Parameter
    type Parameter_Access is access Parameter_Type'Class;
    type Parameter_Record;                 -- Pointer auf Parameter-Satz
    type Record_Access is access Parameter_Record;
    type Parameter_Record                  -- Parameter-Satz
    is   record                            -- enthaelt
             Para: Parameter_Access;       --   Pointer auf Parameter
             Next: Record_Access;          --   Pointer auf naechst.P-Satz
    end record;

    type Tuple_Header                      -- Tupel-Header
    is   record                            -- enthaelt
             First: Record_Access;         --   Pointer auf 1. P-Satz
             Last: Record_Access;          --   Pointer auf letzten
    end record;
                                           -- der Pointer auf den Header
    type Tuple_Access is access all Tuple_Header;

    type Tuple_Type                        -- Das Tupel (=Liste Par.-Saetze)
    is   new Controlled
         with record                       -- erzeugt den Header, der
                                           -- am Anfang auf nix zeigt.
             Header: Tuple_Access
                 := new Tuple_Header'(First => null, Last => null);
    end record;

                                           -- Der Iterator verwaltet das:
    type Iterator_Type                     -- Er besteht aus:
    is   record
             Header: Tuple_Access;         -- dem Pointer auf den Header
             Current: Record_Access;       -- dem Pointer, der auf den
                                           --   akt. Parameter-Satz zeigt
    end record;

    procedure Adjust (T: in out Tuple_Type);

end Tuple;
