-- Realisierung der Tupelraum-Funktionen

with stringpack; use stringpack;

package body Tuplespace is
    -- First: liefert erstes Tupel im Tupelraum
    -- Uebergabe: Tupelraum
    -- Rueckgabe: Iterator mit Tupelraum-Header und Zeiger auf 1. Parameter
    function First (T: Tuplespace_Type) return Iterator_Type is
    begin
        return (T.Header, T.Header.First);
    end First;

    -- Succ:  liefert naechstes Tupel
    -- Uebergabe: Iterator mit Tupelraum-Header und Zeiger auf aktuelles Tupel
    -- Rueckgabe: Iterator mit Tupelraum-Header und Zeiger auf naechstes Tupel
    function Succ (I: Iterator_Type) return Iterator_Type is
    begin
        return (I.Header, I.Current.Next);
    end Succ;

    -- Prev:  liefert voriges Tupel
    -- Uebergabe: Iterator mit Tupelraum-Header und Zeiger auf aktuelles Tupel
    -- Rueckgabe: Iterator mit Tupelraum-Header und Zeiger auf voriges Tupel
    function Prev (I: Iterator_Type) return Iterator_Type is
    begin
        return (I.Header, I.Current.Prev);
    end Prev;

    -- Value: liefert das Tupel des aktuellen Satzes
    -- Uebergabe: Iterator mit Tupelraum-Header und Zeiger auf aktuellen Tupel-Satz
    -- Rueckgabe: Tupel des aktuellen Tupel-Satzes
    function Value (I: Iterator_Type) return Tuple_Type is
    begin
        return I.Current.Tup.all;
    end Value;

    -- End_Of_Tupelraum: zeigt an, ob das letzte Tupel erreicht ist
    -- Uebergabe: Iterator mit Tupelraum-Header und Zeiger auf aktuelles Tupel
    -- Rueckgabe: Boolean: true oder false
    function End_Of_Tuplespace (I: Iterator_Type) return boolean is
    begin
        return I.Current = null;
    end;

    -- Insert_First: fuegt vorne an die Tupel-Liste einen Tupel an
    -- Uebergabe: Tupelraum (= Tupel-Liste) und den neuen Tupel
    procedure Insert_First (T: Tuplespace_Type; P: Tuple_Type) is
        I: Iterator_Type;
    begin
        T.Header.First := new Tuple_Record'
            (null, new Tuple_Type'(P), T.Header.First);
        I := First(T);
        if   I.Current.Next /= null
        then I.Current.Next.Prev := I.Current;
        end if;
    end Insert_First;

    -- Delete: loescht den aktuellen Tupel aus der Tupel-Liste
    -- Uebergabe: der Iterator (= Tupelliste + aktuellem Tupel)
    procedure Delete (I: Iterator_Type) is
    begin
        if I.Current.Prev /= null then
           I.Current.Prev.Next := I.Current.Next;
        else
           I.Header.First := I.Current.Next;
           -- I.Current.Next.Prev := null;
        end if;
        if I.Current.Next /= null then
           I.Current.Next.Prev := I.Current.Prev;
        else
           -- I.Current.Prev.Next := null;
           null;
        end if;

    end Delete;

    -- Find_Matching: Sucht Matchendes Tupel T im Tupelraum
    -- Uebergabe: Tupelraum, Tupel
    -- Rueckgabe: Iterator auf das Tupel
    function Find_Matching (TS: Tuplespace_Type; T: Tuple_Type)
    return   Iterator_Type
    is
       I: Iterator_Type;
    begin
       I := First(TS);
       while not End_Of_Tuplespace (I)
       loop
          if Match (T, Value(I)) then
             -- print ("Match!");
             return I;
          end if;
          -- print ("Noch kein Match!");
          I := Succ (I);
       end loop;
       -- print ("Kein Match!");
       -- Nichts gefunden --> Exception ausloesen
       raise No_Tuple_Matches;
    end;


end Tuplespace;
